rm(list = ls())
options(scipen = 10)
pcName <- tolower(Sys.info()[["nodename"]])
if (pcName == "kogoly-laptop") {
  pathToDropbox <- "C:/Users/kogoly/"
} else if (pcName == "konstantins-macbook-air.local") {
  pathToDropbox <- "~/"
} else stop("Dunno where all files are!")
# pathToDropbox <- "~/"
pathToData <- file.path(pathToDropbox,
                        "Dropbox/Writing/RandomizationPaper/")
# setwd(pathToData)
library("tidyverse")
library("assertr")
library("readr")
library("stringr")
library("digest") # to compute md5 hashes
library("Rmpfr") # to convert md5 to numbers
library("bit64")
library("ggthemes")
# print(listOfFiles)
source(file.path(pathToData, "code/FunctionsLibrary.R"))
###############################################################################
# Starting fusion data
load(file.path(pathToData, "data/FusionDataUsaProductIdsAndAttributes.Rda"))
kQuantilesToCompute <- seq(from = 0, to = 1, by = 0.05)
kAllHashes <- c("md5", "sha1", "crc32", "sha256", "sha512", "xxhash32",
                "xxhash64", "murmur32") # total of 8
kHashAlgoToUse <- kAllHashes[[5]] # fifth is borderline fail, so is 8
kHashesUsedEventually <- c(kAllHashes[[1]], kAllHashes[[5]])
print(kHashAlgoToUse)
###############################################################################
PerformAnalysis <- function(dataFrame, hashColname, quantilesToCompute,
                            hash = kHashAlgoToUse) {
  ksResults <- PerformKSTest(dataFrame[[hashColname]], quantilesToCompute)
  chiSqResults <- PerformChiSquaredTest(dataFrame[[hashColname]])
  result <- data.frame(hashFunction = hash,
                       rbind(ksResults, chiSqResults),
                       stringsAsFactors = FALSE)
  result
} # end PerformAnalysis()
###############################################################################
HashDominicsData <- function(dataPath, idColname, hashColname, hashFunc,
                             quantilesToUse) {
  rawDominicsData <- LoadDominicsData(dataPath)
  cleanDominicsData <- CleanDominicsData(rawDominicsData)
  hashedDominicsData <- ConstructHashes(cleanDominicsData, idColname,
                                        hashColname, hashFunc)
  hashedDominicsData
} # end HashDominicsData()
###############################################################################
HashFusionData <- function(dataPath, idColname, hashColname, hashFunc,
                           quantilesToUse) {
  requireNamespace("dplyr", quietly = TRUE)
  rawFusionData <- readRDS(dataPath)
  cleanFusionData <- dplyr::select_(rawFusionData, idColname)
  hashedFusionData <- ConstructHashes(cleanFusionData, idColname
                                      , hashColname, hashFunc)
  hashedFusionData
} # end HashFusionData()
###############################################################################

dominicsUpcs <- LoadDominicsData(dataDirPath = file.path(pathToData,
                                                         "data/dominics"))
cleanUpcData <- CleanDominicsData(dominicsUpcs)
upcDataWithHashes <- ConstructHashes(cleanUpcData, idColumnName = "upc",
                                     hashColumnName = "upcHash",
                                     hash = kHashAlgoToUse)
ksResults <- PerformKSTest(upcDataWithHashes[["hashScaled"]],
                           kQuantilesToCompute)
chiSqResults <- PerformChiSquaredTest(upcDataWithHashes[["hashScaled"]])
dominicsResults <- rbind(ksResults, chiSqResults)

###############################################################################
###############################################################################
###############################################################################
###############################################################################

kColNames <- c('model_no', "modelHash")

fusionProductsWithMd5 <- dataForPaper %>%
  select(model_no, formfactor, product_family) %>%
  mutate(
    formfactor = tolower(formfactor)
    , product_family = tolower(product_family)
  )

fusionProductsWithMd5 <- ConstructHashes(fusionProductsWithMd5,
                                         idColumnName = "model_no",
                                         hashColumnName = "modelHash",
                                         hash = kHashAlgoToUse)

### Doing formal tests
ksResults <- PerformKSTest(fusionProductsWithMd5[["hashScaled"]],
                           kQuantilesToCompute)
chiSqResults <- PerformChiSquaredTest(fusionProductsWithMd5[["hashScaled"]])
fusionResults <- rbind(ksResults, chiSqResults)

print(dominicsResults)
print(fusionResults)
print(kHashAlgoToUse)
print(upcDataWithHashes[1,], width = Inf)


###############################################################################
Bd <- 2000
set.seed(1)
resultsByAlgo <- vector("list", length(kHashesUsedEventually))
for (i in seq_along(kHashesUsedEventually)) {
  # i = 1
  hashAlgo <- kHashesUsedEventually[[i]]
  dominicsDataHashed <- HashDominicsData(
    dataPath = file.path(pathToData, "data/dominics"), idColname = "upc"
    , hashColname = "hashScaled", quantilesToUse = kQuantilesToCompute
    , hashFunc = hashAlgo)
  upcResults <- vector("list", Bd)
  Nd <- NROW(dominicsDataHashed)
  for (b in seq_len(Bd)) {
    idx <- sample.int(Nd, size = Nd, replace = TRUE)
    bootstrapData <- dominicsDataHashed[idx, ]
    result <- PerformAnalysis(bootstrapData, hashColname = "hashScaled"
                              , quantilesToCompute = kQuantilesToCompute
                              , hash = hashAlgo)
    upcResults[[b]] <- result %>%
      mutate(
        nObs = NULL
        , boostrapDraw = b
        , testName = str_replace_all(testType, "-Exact", "")
        , testType = NULL
        , key = str_replace_all(key, "test", "")
      )
    if (b %% 100 == 0) print(b)
  } # end for each bootstrap
  resultsByAlgo[[i]] <- bind_rows(upcResults)
} # end for each hash algo
dominicsBootstrap <- bind_rows(resultsByAlgo)
pryr::object_size(dominicsBootstrap)
str(dominicsBootstrap)
kTestName <- "ChiSq"
kTestName <- "KS"
dominicsBootstrap %>%
  filter(key == "PValue") %>%
  qplot(value, data = .) +
  facet_grid(hashFunction ~ testName) +
  ggtitle(paste(kTestName, kHashAlgoToUse))

Round <- function(x, f, ...) round(f(x, ..., na.rm = TRUE), digits = 2)

summz <- dominicsBootstrap %>%
  filter(key == "PValue") %>%
  group_by(hashFunction, testName) %>%
  summarize(
    min      = Round(value, min)
    , p0p025 = Round(value, quantile, prob = 0.025)
    , p0p05  = Round(value, quantile, prob = 0.05)
    , p0p10  = Round(value, quantile, prob = 0.10)
    , p0p25  = Round(value, quantile, prob = 0.25)
    , p0p50  = Round(value, quantile, prob = 0.50)
    , p0p75  = Round(value, quantile, prob = 0.75)
    , p0p90  = Round(value, quantile, prob = 0.90)
    , p0p95  = Round(value, quantile, prob = 0.95)
    , p0p975 = Round(value, quantile, prob = 0.975)
    , max    = Round(value, max)
    , avg    = Round(value, mean)
  ) %>%
  ungroup() %>%
  gather(statistic, value, min:avg) %>%
  mutate(
    hash_plus_test = str_c(hashFunction, testName, sep = "_")
    , hashFunction = NULL
    , testName     = NULL
  ) %>%
  # select(-hashFunction, -testName) %>%
  spread(hash_plus_test, value)
summz

distribz <- dominicsBootstrap %>%
  filter(key == "PValue") %>%
  mutate(
    hash_plus_test = str_c(hashFunction, testName, sep = "_")
    , hashFunction = NULL
    , testName     = NULL
  ) %>%
  spread(hash_plus_test, value)

for (n in names(distribz)[-c(1, 2)]) {
  print(str_c(n, sum(distribz[[n]] <= 0.05)/NROW(distribz), sep = " "))
}

dominicsPlotData <- dominicsBootstrap %>%
  filter(key == "PValue") %>%
  mutate(
    testName = ifelse(testName == "KS", "Kolmogorov-Smirnoff",
                      ifelse(testName == "ChiSq", "Chi-squared", "???"))
    , hashFunction = toupper(hashFunction)
  )
dominicsResultsNoBootstrap <- data.frame(
  hashFunction = rep(c("MD5", "SHA512"), 2)
  , testName = rep(c("Kolmogorov-Smirnoff", "Chi-squared"), each = 2)
  , PValues    = c(.884, .396, .814, .052)
  , stringsAsFactors = FALSE
)
dominicsResultsNoBootstrap
# making the plot
setEPS()
postscript(file.path(pathToData, "plots/Dominics-Plot.eps"),
           paper = "letter")

ggplot(data = dominicsPlotData, aes(x = value)) +
  geom_histogram(aes(group = hashFunction, fill = hashFunction)
                 , bins = 20, position = "dodge") +
  facet_grid(testName ~ .) +
  geom_vline(
    data = dominicsResultsNoBootstrap
    , mapping = aes(xintercept = PValues, color = hashFunction)
  ) +
  guides(color = FALSE) + # suppress secondary legend
  labs(
    x = "Test P-Values"
    , y = "Counts"
    , fill = "Hash Function"
    , title = str_c("Distribution of P-Values Across Tests and Hashing ",
                    "Algorithms For Dominick's Data")
  ) +
  theme_bw() + scale_fill_grey() + scale_color_grey()
dev.off()


###############################################################################
###############################################################################
###############################################################################
###############################################################################
###############################################################################

Bd <- 2000
set.seed(1)
fusion_resultsByAlgo <- vector("list", length(kHashesUsedEventually))
for (i in seq_along(kHashesUsedEventually)) {
  # i = 1
  hashAlgo <- kHashesUsedEventually[[i]]
  fusionDataHashed <- HashFusionData(
    dataPath = file.path(pathToData,
                         "data/FusionDataUsaProductIdsAndAttributes.rds")
    , idColname = "model_no", hashColname = "hashScaled"
    , quantilesToUse = kQuantilesToCompute, hashFunc = hashAlgo)
  upcResults <- vector("list", Bd)
  Nd <- NROW(fusionDataHashed)
  for (b in seq_len(Bd)) {
    idx <- sample.int(Nd, size = Nd, replace = TRUE)
    bootstrapData <- fusionDataHashed[idx, ]
    result <- PerformAnalysis(bootstrapData, hashColname = "hashScaled"
                              , quantilesToCompute = kQuantilesToCompute
                              , hash = hashAlgo)
    upcResults[[b]] <- result %>%
      mutate(
        nObs = NULL
        , boostrapDraw = b
        , testName = str_replace_all(testType, "-Exact", "")
        , testType = NULL
        , key = str_replace_all(key, "test", "")
      )
    if (b %% 100 == 0) print(b)
  } # end for each bootstrap
  fusion_resultsByAlgo[[i]] <- bind_rows(upcResults)
} # end for each hash algo
fusionBootstrap <- bind_rows(fusion_resultsByAlgo)
pryr::object_size(fusionBootstrap)
str(fusionBootstrap)
kTestName <- "ChiSq"
kTestName <- "KS"
fusionBootstrap %>%
  filter(key == "PValue") %>%
  qplot(value, data = .) +
  facet_grid(hashFunction ~ testName) +
  ggtitle(paste(kTestName, kHashAlgoToUse))

# Round <- function(x, f, ...) round(f(x, ..., na.rm = TRUE), digits = 2)

fusion_summz <- fusionBootstrap %>%
  filter(key == "PValue") %>%
  group_by(hashFunction, testName) %>%
  summarize(
    min      = Round(value, min)
    , p0p025 = Round(value, quantile, prob = 0.025)
    , p0p05  = Round(value, quantile, prob = 0.05)
    , p0p10  = Round(value, quantile, prob = 0.10)
    , p0p25  = Round(value, quantile, prob = 0.25)
    , p0p50  = Round(value, quantile, prob = 0.50)
    , p0p75  = Round(value, quantile, prob = 0.75)
    , p0p90  = Round(value, quantile, prob = 0.90)
    , p0p95  = Round(value, quantile, prob = 0.95)
    , p0p975 = Round(value, quantile, prob = 0.975)
    , max    = Round(value, max)
    , avg    = Round(value, mean)
  ) %>%
  ungroup() %>%
  gather(statistic, value, min:avg) %>%
  mutate(
    hash_plus_test = str_c(hashFunction, testName, sep = "_")
    , hashFunction = NULL
    , testName     = NULL
  ) %>%
  # select(-hashFunction, -testName) %>%
  spread(hash_plus_test, value)
fusion_summz

fusion_distribz <- fusionBootstrap %>%
  filter(key == "PValue") %>%
  mutate(
    hash_plus_test = str_c(hashFunction, testName, sep = "_")
    , hashFunction = NULL
    , testName     = NULL
  ) %>%
  spread(hash_plus_test, value)

for (n in names(fusion_distribz)[-c(1, 2)]) {
  print(str_c(n, sum(fusion_distribz[[n]] <= 0.05) /
                NROW(fusion_distribz), sep = " "))
}

fusionPlotData <- fusionBootstrap %>%
  filter(key == "PValue") %>%
  mutate(
    testName = ifelse(testName == "KS", "Kolmogorov-Smirnoff",
                      ifelse(testName == "ChiSq", "Chi-squared", "???"))
    , hashFunction = toupper(hashFunction)
  )
fusionResultsNoBootstrap <- data.frame(
  hashFunction = rep(c("MD5", "SHA512"), 2)
  , testName = rep(c("Kolmogorov-Smirnoff", "Chi-squared"), each = 2)
  , PValues    = c(.867, .078, .385, .099)
  , stringsAsFactors = FALSE
)
fusionResultsNoBootstrap
# making the plot
setEPS()
postscript(file.path(pathToData, "plots/PC-Sales-Plot.eps"),
           paper = "letter")

ggplot(data = fusionPlotData, aes(x = value)) +
  geom_histogram(aes(group = hashFunction, fill = hashFunction)
                 , bins = 20, position = "dodge") +
  facet_grid(testName ~ .) +
  geom_vline(
    data = fusionResultsNoBootstrap
    , show.legend = FALSE
    , mapping = aes(xintercept = PValues
                    , color = as.factor(hashFunction))
  ) +
#  guides(color = FALSE) + # suppress secondary legend
  labs(
    x = "Test P-Values"
    , y = "Counts"
    , fill = "Hash Function"
    , title = str_c("Distribution of P-Values Across Tests and Hashing ",
                    "Algorithms For PC Sales Data")
  ) +
  theme_bw() + scale_fill_grey() + scale_color_grey()
dev.off()





